"""
The docker tool provides ``build``, ``push`` and ``publish`` commands.
"""
import argparse
import logging

from menhir.tool import Tool
from menhir.tool_utils import OK, FAIL, NOTHING_TO_DO, tool_env, working_dir

log = logging.getLogger(__name__)


def tool():
    return Docker()


class Docker(Tool):

    def name(arg):
        return "docker"

    def dir_info(tool, path, info):
        from os.path import exists, join
        path = join(path, 'Dockerfile')
        has_dockerfile = exists(path)
        return {
            'project_recognised': has_dockerfile,
            'can_run': has_dockerfile,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        from os.path import exists, join
        from menhir.tool_utils import run_if

        phase_name = args.phase

        dockerfile = join(path, 'Dockerfile')
        repo = info['config'].get('docker', {}).get('repository')

        if not exists(dockerfile):
            log.debug(
                'No Dockerfile %(dockerfile)s',
                {'dockerfile': dockerfile})
            return NOTHING_TO_DO

        run_flag = (
            'changed' not in info or
            info['changed'].get('self') or
            info['changed'].get('dependents')
        )

        if phase_name == 'build':
            with run_if(run_flag, phase_name, path) as flag:
                if flag:
                    return docker_build(path, info, args)
                return OK

        elif phase_name == 'push':
            with run_if(run_flag, phase_name, path) as flag:
                if flag:
                    return docker_push(repo, path, info, args)
                return OK

        elif phase_name == 'publish':
            with run_if(run_flag, phase_name, path) as flag:
                if flag:
                    res = docker_build(path, info, args)
                    if res != OK:
                        return res
                    return docker_push(repo, path, info, args)
                return OK
        else:
            return NOTHING_TO_DO


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Commands to build and push docker images.",
        **kwargs
    )
    parsers = parser.add_subparsers(help="Docker commands", dest='phase')
    parsers.add_parser(
        'build',
        help='Build a docker image from a Dockerfile'
    )
    parsers.add_parser(
        'push',
        help='Push a docker image to a remote repository'
    )
    parsers.add_parser(
        'publish',
        help='Build and push a docker image to a remote repository'
    )
    return parser


def docker_build(path, info, args):
    from menhir.tool_utils import call, package_script
    log.info('Running docker-build in %s', path)

    project_name = info['project-name']

    env = tool_env()
    env['MENHIR_TAG'] = project_name

    with package_script("/tools/docker/docker-build.sh") as f:
        with working_dir(path):
            return call([f.name], env=env,)


def docker_push(repo, path, info, args):
    from menhir.project import branch, image
    from menhir.tool_utils import call, package_script, slugify
    log.info('Running docker-push in %s', path)

    project_name = info['project-name']
    current_branch = branch()
    tag = project_name
    sha_tag = image(info, path)
    if not sha_tag:
        log.error('No remote repository configured to push to.')
        return FAIL
    branch_tag = "%s:%s" % (
        sha_tag.split(':')[0],
        slugify(current_branch, length=40),
    )

    env = tool_env()
    env['MENHIR_TAG'] = tag
    env['MENHIR_BRANCH_TAG'] = branch_tag
    env['MENHIR_SHA_TAG'] = sha_tag

    with package_script("/tools/docker/docker-push.sh") as f:
        with working_dir(path):
            return call([f.name], env=env,)



# menhir configuration
import logging
import os

log = logging.getLogger(__name__)


def project_config(path="menhir-config.yaml", dir=os.getcwd()):
    """Return the project configuration in dir_path."""
    from os.path import abspath, basename, dirname, isabs, join
    from menhir.fileutils import load_yaml
    if not isabs(path) and path != "-":
        path = join(dir, path)
    config = load_yaml(path) or {}
    if 'project-name' not in config:
        config['project-name'] = basename(dirname(abspath(path)))
    config['config'] = config.get('config', {})
    return config



"""Utilities for use in tools.

"""

import logging
import re
from contextlib import contextmanager

log = logging.getLogger(__name__)

OK = {
    'status': 'ok'
}

FAIL = {
    'status': 'fail'
}

NOTHING_TO_DO = {
    'status': 'nothing_to_do'
}

NON_WORD_PATTERN = re.compile(r'\W')


@contextmanager
def package_script(resource_path, resource_package="menhir"):
    """Execute a block of code with the given script from the package.

    Yields a file like object that is the script written onto the filesystem.
    """
    import tempfile
    import pkg_resources
    import stat
    from os import chmod, remove

    script = pkg_resources.resource_string(resource_package, resource_path)
    fname = None
    try:
        with tempfile.NamedTemporaryFile("wb", delete=False) as f:
            fname = f.name
            f.write(script)
        chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
        yield f
    finally:
        remove(fname)


@contextmanager
def working_dir(path):
    """Execute a block of code within the given working dir."""
    import os
    dir = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(dir)


def tool_env():
    """Return the default tool environment dict."""
    import os
    import os.path

    default_pyenv_root = os.path.join(os.getenv("HOME"), ".pyenv")
    env = {
        "HOME": os.getenv("HOME"),
        "PATH": os.getenv("PATH"),
        "PYENV_ROOT": os.getenv("PYENV_ROOT", default_pyenv_root),
    }

    v = os.getenv("AWS_DEFAULT_REGION")
    if v:
        env["AWS_DEFAULT_REGION"] = v

    v = os.getenv("LANG")
    if v:
        env["LANG"] = v

    v = os.getenv("LC_ALL")
    if v:
        env["LC_ALL"] = v

    v = os.getenv("DEBUG")
    if v:
        env["DEBUG"] = v

    return env


def call(cmd, *args, **kwargs):
    """Call a subprocess, returning a menhir result."""
    import subprocess
    res = subprocess.call(cmd, *args, **kwargs)
    if res:
        return FAIL
    return OK


def slugify(s, length=None, replace=NON_WORD_PATTERN):
    s = re.sub(replace, "_", s)
    if length:
        s = s[:length]
    return s


@contextmanager
def run_if(cond, phase_name, path):
    if cond:
        log.info(
            'Running %(phase)s in %(path)s',
            {'phase': phase_name, 'path': path}
        )
        yield True
    else:
        yield False
        log.info(
            'Not running %(phase)s in %(path)s',
            {'phase': phase_name, 'path': path}
        )



"""The dynamodb tool provides the ``put`` task.

The ``put`` task is used to upload json files to DynamoDB tables.  The
files to upload, and the target tables are specified in the project
configuration under the ``dynamodb`` key.

The ``targets`` sub-key, contains a mapping of name to configuration,
where each specifies ``source``, ``args``, ``table``, ``values`` and
``descrypt`` keys.


Source
------

The ``source`` key specifies a pattern to match the source files to
upload.


Table
-----

The ``table`` key specifies the table name to upload to.  Each source
file is uploaded independently to the table.

The table value is a python format string that may contain
interpolations for the args and values specified below.


Arguments
---------

A directory target can be configured with ``args``, a list containing
the names of values it expects to be passed on the command line.


Values
------

The ``values`` key configures a list of names for values to be added
to the top level of the JSON before upload.

``values`` can contain names from ``args`` and the following:

project        the name of the project
branch         the current branch name
branch-slug    the current branch name, sanitised
sha            the current commit sha as a hex string
sha-mod-1024   the current as a decimal modulo 1024

The ``values`` list can also contain items of the form ``name=value``,
which will pass the var ``name`` to terraform, with the value
calculated according to ``value`` from the list above.


Decrypt
-------

The ``decrypt`` key specifies that the files matched by ``source``
should first be decrypted.  The value of the key specifies an
environment variable that holds the decryption key.

"""
import argparse
import logging

from menhir.tool import Tool
from menhir.utils import multi, method
from menhir.tool_utils import OK, FAIL, NOTHING_TO_DO

log = logging.getLogger(__name__)

REMOVE_SUBS_RE = r'%\([\w_]+\)s'


def tool():
    return DynamoDB()


class DynamoDB(Tool):

    def dir_info(tool, path, info):
        """Does a glob match on the source patterns.

        This will not be 100% accurate, but will not give false
        negatives.
        """
        import re
        from glob import glob
        from os.path import join
        files = None
        targets = info.get('dynamodb', {}).get('targets')
        if targets:
            for k, v in targets.items():
                source = v.get('source', 'config.*.json')
                source = re.sub(REMOVE_SUBS_RE, '*', source)
                files = glob(join(path, source))
        return {
            'project_recognised': False,
            'can_run': bool(files),
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        return task(path, info, args)


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Commands to upload JSON to dynamodb table.",
        **kwargs
    )
    parsers = parser.add_subparsers(help="DynamoDB commands", dest='phase')
    p = parsers.add_parser(
        'put',
        help='Put JSON files to dynamodb'
    )
    p.add_argument('target', nargs=1)
    p.add_argument('args', nargs="*")
    return parser


@multi
def task(path, info, args):
    return args.phase


@method(task, 'put')
def task_put(path, info, args):
    if not (
            'changed' not in info or
            info['changed'].get('self') or
            info['changed'].get('dependents')
    ):
        return NOTHING_TO_DO
    config = info.get('config', {}).get('dynamodb', {})
    targets = config.get('targets', {})
    target = args.target[0]
    spec = targets.get(target)
    print('config', config, 'targets', targets)
    print('target', target, 'spec', spec)
    if spec is None:
        return NOTHING_TO_DO

    res = put(path, info, target, args.args, spec)
    if res != OK:
        return res
    return OK


def put(path, info, target, args, spec):
    from glob import glob
    from os import getenv
    from os.path import join
    import boto3
    from menhir.tool_config import aliased_value_array

    print('put')
    args_names = spec.get('args', [])
    args_dict = dict(zip(args_names, args))

    values = spec.get('values', [])
    values_map = {
        k: v for k, v in aliased_value_array(values, info, path, args_dict)
    }

    values_and_args = args_dict.copy()
    values_and_args.update(values_map)

    source = spec.get('source', 'config.json')
    source = source % values_and_args

    table_name = spec.get('table', 'config.%(project)s')
    table_name = table_name % values_and_args

    decrypt = spec.get('decrypt')
    if decrypt:
        decrypt = getenv(decrypt)
        if not decrypt:
            log.error(
                'Decryption key %s specified, but not present',
                spec.get('decrypt')
            )
            return FAIL

    dynamodb = boto3.resource('dynamodb')
    table = dynamodb.Table(table_name)
    files = glob(join(path, source))

    print('files', files)
    if not files:
        log.error(
            'No files match %(source)s for target %(target)s',
            {'source': source, 'target': target}
        )
        return FAIL

    for file in files:
        log.info('Uploading %s to %s', file, table_name)
        put_file(table, values_map, file, decrypt)

    return OK


def put_file(table, values_map, file, decrypt):
    from json import loads
    from menhir.fileutils import load_json, load_encrypted

    if decrypt:
        s = load_encrypted(file, decrypt)
        data = loads(s)
    else:
        data = load_json(file)
    data.update(values_map)
    log.info('Uploading %s', data)

    table.put_item(Item=data)



"""Functions for working with tool configuration.

"""
from menhir.utils import multi, method


def value_array(value_names, info, path, extra_values):
    """Return a list of values for the list of ``value_names``.

    ``value_names`` are first looked up in extra_values.  If not
    specified there, then they are evaluated with ``config_value``.
    """
    def lookup_value(x):
        return lookup_config_value(x, info, path, extra_values)

    return list(map(lookup_value, value_names))


def aliased_value_array(value_names, info, path, extra_values):
    """Return a list of tuples for the list of ``value_names``.

    Each item in ``value_names`` is either a simple config value name,
    or is an ``alias=name`` specification.

    Each returned tuple contains the ``alias`` (defaults to config
    value name) and it's value.

    Value names are first looked up in extra_values.  If not specified
    there, then they are evaluated with ``config_value``.
    """
    def lookup_value(x):
        if '=' in x:
            [n, v] = x.split('=')
        else:
            n = v = x
        return (n, lookup_config_value(v, info, path, extra_values))

    return list(map(lookup_value, value_names))


def lookup_config_value(x, info, path, extra_values):
    value = extra_values.get(x)
    if value is None:
        value = config_value(x, info, path)
    return value


@multi
def config_value(name, info, path):
    """Return a configuration value for passing to terraform."""
    return name


@method(config_value, 'project')
def config_value_project(name, info, path):
    return info['project-name']


@method(config_value, 'branch')
def config_value_branch(name, info, path):
    from menhir.project import branch
    return branch()


@method(config_value, 'branch-slug')
def config_value_branch_slug(name, info, path):
    from menhir.project import branch
    from menhir.tool_utils import slugify
    return slugify(branch(), length=22)


@method(config_value, 'sha-mod-1024')
def config_value_sha_mod_1024(name, info, path):
    from menhir.project import sha
    return int(sha(), 16) % 1024


@method(config_value, 'sha')
def config_value_sha(name, info, path):
    from menhir.project import sha
    return sha()


@method(config_value, 'image')
def config_value_image(name, info, path):
    from menhir.project import image
    return image(info, path)



"""Tool interface for tool implementations.

A tool is implemented as an opaque class.

"""
import logging

log = logging.getLogger(__name__)


def missing_impl(tool, f):
    raise Exception(
        'incorrect tool implementation',
        'Tool implementation %s is missing a "%s" implementation' % (
            type(tool),
            f,
        )
    )


class Tool(object):
    def __str__(self):
        return "<%s>" % type(self).__name__

    def __repr__(self):
        return "<%s>" % type(self).__name__

    def name(self):
        return type(self).__name__.lower()

    def is_higher_order(self):
        return False

    def dir_info(tool, path, info):
        """Return info about using the tool in the specified path."""
        missing_impl(tool, Tool.dir_info.__name__)

    def dependencies(tool, path):
        """Return a list of dependency prefixes for the project at path."""
        missing_impl(tool, Tool.dependencies.__name__)

    def configure_parser(tool, parser):
        """Configure arg parser for the tool options and arguments."""
        missing_impl(tool, Tool.add_arg_parser.__name__)

    def execute_tool(tool, path, info, args):
        """Execute a build phase."""
        missing_impl(tool, Tool.execute_tool.__name__)


def load_tool(tool_name):
    """Load the tool implementation for the given tool name."""
    tool = \
        require_tool(tool_name) or \
        require_tool('menhir.tools.%s' % tool_name)
    if not tool:
        raise Exception(
            'configuration_error',
            'Tool not found: "%s"' % tool_name)
    return tool


def require_tool(tool_path):
    from importlib import import_module
    try:
        m = import_module(tool_path)
        return m.tool()
    except Exception as e:
        log.debug('Failed to import: %s, "%s"', tool_path, e)


def default_tools():
    """Return a list of the default tool names."""
    from menhir.tools import all_tools
    return all_tools



"""Tool to print menhir's configuration and state dict.

This can be used to inspect what the configuration a previous tool
produces, or the next tool sees.

It can output in python pprint, json or yaml formats.
"""
import argparse
import logging
import pprint

import json
import yaml

from menhir.tool import Tool
from menhir.tool_utils import OK, FAIL

log = logging.getLogger(__name__)


def tool():
    return Pprint()


class Pprint(Tool):

    def dir_info(tool, path, info):
        return {
            'project_recognised': False,
            'can_run': True,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Pprint the menhir config."""
        if args.format == 'pprint':
            pprint.pprint(info, width=80)
        elif args.format == 'yaml':
            print(yaml.dump(
                info,
                width=80,
                indent=2,
                default_flow_style=False,
            ))
        elif args.format == 'json':
            print(json.dumps(info, sort_keys=True, indent=4))
        else:
            log.error('Invalid output format: "%s', args.format)
            return FAIL
        return OK


def parser(**kwargs):
    """Configure arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="Pretty print the menhir configuration.",
        **kwargs
    )
    parser.add_argument(
        "--format",
        default='yaml',
        choices=['json', 'pprint', 'yaml'],
        help="Output format.",
    )
    return parser
